#!/bin/bash
#
# Simple toolkit to manage Samba suite v3+ standalone fileserver on Debian.
# Part of samba-manager package. Sorry about my terrible English! :-)

# Including headers
HEADERS="/etc/default/smbm"
MSG_SMBM_NOHEADER="Header file $HEADERS is not found - terminating."
if [ ! -r "$HEADERS" ]; then echo -e "$MSG_SMBM_NOHEADER" >&2; exit 1; fi
. $HEADERS

#
# Functions
#

# smbm_adduser S/W_username [password [Linux_username]]
# Adds a Samba/Windows user with corresponding Linux user.
function smbm_adduser {

    local domain=`$HOSTNAME | $AWK '{ print toupper($0) }'`
    local result

    # S/W username is mandatory and formally checked.
    local winusername="$1"; shift
    if [ -z "$winusername" ]; then log $MSG_SMBM_ALL_NOWINUSER; return $EXIT_ERR; fi
    if [ ! -z "$(echo -e "$winusername" | $SED "s/$MASKWINUSER//")" ]; then log $MSG_SMBM_ALL_NOWINUSER; return $EXIT_ERR; fi
    # Error if already exists.
    local winuserfound=$(getwinuser "$winusername")
    if [ -n  "$winuserfound" ]; then log "$MSG_SMBM_ALL_ENTITY_EXISTS $winuserfound"; return $EXIT_ERR; fi

    # Password is optional, if empty, random password used.
    # Password is formally checked also.
    local password="$1"; shift
    if [ -z "$password" ]; then password=`$MAKEPASSWD`; fi
    if [ ! -z "`echo "$password" | $SED "s/$MASKPASS//"`" ]; then log "$MSG_SMBM_ALL_WRONGPWD"; return $EXIT_ERR; fi

    # User entity planning.
    # Get next available SID-RID and computes corresponding Linux UID.
    local localsid=$($NET getlocalsid | $SED "s/.*: //")
    local nextrid=`getnextrid`
    if [ -z "$nextrid" ]; then log "$MSG_SMBM_ALL_OUTOFRID"; return $EXIT_ERR;fi
    local linuid=$(( LINUIDBASE + nextrid ))
    # Making a corresponding Linux username.
    local linusername="$1"; shift
    # If unspecified, generates an unique RID mapped Linux username.
    if [ -z "$linusername" ]; then linusername="$LINUSER$linuid"; fi
    # Formal check, giving up on failure.
    if [ ! -z "$(echo -e "$linusername" | $SED "s/$MASKLINUSER//")" ]; then log $MSG_SMBM_ALL_NOLINUSER; return $EXIT_ERR; fi
    # Check for existence of Linux user given/planned.
    local pwdentry=`getlinuser "$linusername"`
    local smbentry=`getentity -lu "$linusername" 2>/dev/null`
    if [ -n "$pwdentry" ]; then
        # Existing user with existing Samba entry isn't acceptable.
        if [ -n "$smbentry" ]; then log "$MSG_SMBM_ALL_ENTITY_EXISTS $smbentry"; return $EXIT_ERR; fi
        # Existing user with no Samba entry causes many violations:
        # - name isn't RID mapped;
        linusername=$(echo "$pwdentry" | $CUT -f1 -d ':')
        # - UID isn't equal to LINUIDBASE+RID;
        linuid=$(echo "$pwdentry" | $CUT -f2 -d ':')
        # - finally Gecos field isn't containing Samba/Windows username.
        # Warning about violations, but continued.
        log "$MSG_SMBM_ALL_ENTITY_VIOLATES u:$linusername:$linuid:$winusername:$localsid-$nextrid"
    else
        # New username needs formal check.
        if [ ! -z "$(echo -e "$linusername" | $SED "s/$MASKLINUSER//")" ]; then 
	    log $MSG_SMBM_ALL_NOLINUSER; return $EXIT_ERR; 
	fi
	# Specified user name maybe causes a violation:
	# - name isn't RID mapped.
	if [ ! "$linusername" = "$LINUSER$linuid" ]; then
    	    # Warning about violation, but continued.
    	    log "$MSG_SMBM_ALL_ENTITY_VIOLATES u:$linusername:$linuid:$winusername:$localsid-$nextrid"
	fi
    fi
    # Here we've a planned user entity and a suitable $password.
    log "$MSG_SMBM_ALL_ENTITY_CREATING u:$linusername:$linuid:$winusername:$localsid-$nextrid"
    log "$MSG_SMBM_ALL_PW_USED $password"

    # Creating Linux user if isn't exists.
    if [ -z "$pwdentry" ]; then
	# Adding a special Linux user with no shell and no valid password,
	# Samba-standard home dir and with $winusername in Gecos field.
	$USERADD -u $linuid -g "$SMBDEFGROUP" \
	         -d "$SMBROOT/$SMBDIR_HOMES/$linusername" \
	         -M -c "$winusername"  -s "/bin/false" \
	         "$linusername"; result=$?
	# On failure giving up.
	if [ ! $result -eq 0 ]; then
	    log "$MSG_SMBM_CONF_LINUSER_FAILED $linusername!"
	    log "$MSG_SMBM_ALL_ERROR $result!"; return $result
	fi
    fi
    # Creating home, ssh  and profile directories also.
    if [ ! -d "$SMBROOT/$SMBDIR_HOMES/$linusername" ]; then
	log $(mkdir -v -m 750 "$SMBROOT/$SMBDIR_HOMES/$linusername" 2>/dev/null)
	chown $linusername "$SMBROOT/$SMBDIR_HOMES/$linusername" 2>/dev/null
    fi
    if [ ! -d "$SMBROOT/$SMBDIR_HOMES/$linusername/$SSHKEYSDIR" ]; then
	log $(mkdir -v -m 700 "$SMBROOT/$SMBDIR_HOMES/$linusername/$SSHKEYSDIR" 2>/dev/null)
	chown $linusername "$SMBROOT/$SMBDIR_HOMES/$linusername/$SSHKEYSDIR" 2>/dev/null
    fi
    if [ ! -d "$SMBROOT/$SMBDIR_PROFILES/$linusername" ]; then
	log $(mkdir -v -m 750 "$SMBROOT/$SMBDIR_PROFILES/$linusername" 2>/dev/null)
	chown $linusername "$SMBROOT/$SMBDIR_PROFILES/$linusername" 2>/dev/null
    fi
    # Creating a keypair for Samba tunneling.
    if [ ! -f "$SMBROOT/$SMBDIR_HOMES/$linusername/$SSHKEYSDIR/$SSHKEYSFILE" ]; then
	local keyname="${linusername}_AT_$($HOSTNAME --fqdn)"
	# Generating a keypair with current Samba password as passphrase.
	keygen "$SMBROOT/$SMBDIR_HOMES/$linusername/$keyname" "$password" "$winusername $keyname"
	if [ -f "$SMBROOT/$SMBDIR_HOMES/$linusername/$keyname" ]; then
	    # On success adjusting permissions and leaving private key in home directory.
	    chown $linusername "$SMBROOT/$SMBDIR_HOMES/$linusername/$keyname" 2>/dev/null
	    chmod 600 "$SMBROOT/$SMBDIR_HOMES/$linusername/$keyname" 2>/dev/null
	    chown $linusername "$SMBROOT/$SMBDIR_HOMES/$linusername/$keyname.ppk" 2>/dev/null
	    chmod 600 "$SMBROOT/$SMBDIR_HOMES/$linusername/$keyname.ppk" 2>/dev/null
	    # Moving public key into .ssh/authorized_keys2 file with restrictions.
	    echo -ne "$SSHKEYOPTS " \
	        > "$SMBROOT/$SMBDIR_HOMES/$linusername/$SSHKEYSDIR/$SSHKEYSFILE"
	    cat "$SMBROOT/$SMBDIR_HOMES/$linusername/$keyname.pub" \
	        >> "$SMBROOT/$SMBDIR_HOMES/$linusername/$SSHKEYSDIR/$SSHKEYSFILE"
	    chown $linusername "$SMBROOT/$SMBDIR_HOMES/$linusername/$SSHKEYSDIR/$SSHKEYSFILE"
	    chmod 400 "$SMBROOT/$SMBDIR_HOMES/$linusername/$SSHKEYSDIR/$SSHKEYSFILE"
	    chattr +i "$SMBROOT/$SMBDIR_HOMES/$linusername/$SSHKEYSDIR/$SSHKEYSFILE"
	    # Removing original public key.
	     rm "$SMBROOT/$SMBDIR_HOMES/$linusername/$keyname.pub" 2>/dev/null
	    # Reporting success.
	    log "$MSG_SMBM_USER_KEYGEN_OK $keyname"
	else
	    # Warning on failure.
	    log "$MSG_SMBM_USER_KEYGEN_FAILED"
	fi
    fi
    # Existing Linux user.

    # Creating Samba/Windows user with password given.
    echo -e "$password\n$password\n" |\
    $PDBEDIT -a -t -u $linusername -U "$localsid-$nextrid" \
             -h "\\\\$domain\\$SMBDIR_HOMES\\$linusername" \
             -p "\\\\$domain\\$SMBDIR_PROFILES\\$linusername" \
	     -D "$SMBHOMEDRIVE" -f "$winusername" >/dev/null 2>&1; result=$?
    # On failure giving up.
    if [ ! $result -eq 0 ]; then
        log "$MSG_SMBM_ALL_SMBUSER_FAILED $winusername!"
	log "$MSG_SMBM_ALL_ERROR $result!"; return $result
    fi
    # Mapping Samba/Windows username to Linux username.
    usermap "$linusername:$winusername"; result=$?

    # Final check.
    entity=$(getentity -lu "$linusername"); result=$?
    if [ ! $result -eq 0 ]; then
	# Reports success.
	log "$MSG_SMBM_ALL_ENTITY_CREATED $entity"
    fi

    return
}

# smbm_deluser

# smbm_moduser S/W_username [ password [ new_S/W_username ]]
# Modifies an existing Samba/Windows user account (actually changes password).
function smbm_moduser {

    local message
    local result=0

    # S/W username is mandatory and must be existing.
    local winusername="$1"; shift
    if [ -z "$winusername" ]; then log $MSG_SMBM_ALL_NOWINUSER; return $EXIT_ERR; fi
    local userentity=$(getentity -wu "$winusername" 2>/dev/null)
    if [ -z "$userentity" ]; then log "$MSG_SMBM_ALL_NOWINUSER"; return $EXIT_ERR; fi
    # Getting corresponding Linux username.
    local linusername=$(echo -e "$userentity" | $CUT -f2 -d ':')

    # Password is optional, but formally checked.
    local password="$1"; shift
    if [ ! -z "`echo "$password" | $SED "s/$MASKPASS//"`" ]; then log "$MSG_SMBM_ALL_WRONGPWD"; return $EXIT_ERR; fi

    # Password change if requested.
    if [ -n "$password" ]; then
	message=$(echo -e "$password\n$password" | $PDBEDIT -a -t -u "$linusername" >/dev/null)
	result=$?
	# Logging message if any.
	if [ -n "$message" ]; then log "$message"; fi
	# Reporting success or failure.
	if [ $result -eq 0 ]; then
	    log "$MSG_SMBM_ALL_PW_CHANGED $userentity ($($PDBEDIT -wL $linusername))"
	else
	    log "$MSG_SMBM_ALL_PW_CHANGEFAILED $userentity ($($PDBEDIT -wL $linusername))"
	    log "$MSG_SMBM_ALL_ERROR $result!"; return $result
	fi
    fi
    # Password change request handled.

    return
}

# smbm_listuser [ -lu|-wu regexp_pattern ]
# Listing entities with matching Linux OR Samba/Windows usernames to pattern.
function smbm_listuser {

    local pdbuser
    local linusername
    local winusername

    # Getting optional option switch.
    local option=`echo $1 | $SED -n "s/^-\+//p"`
    # If option given a pattern is mandatory.
    local pattern=""
    if [ -n "$option" ]; then
	shift; pattern="$1"
	# Giving up on missing pattern.
	if [ -z "$pattern" ]; then log "$MSG_SMBM_LIST_NOPATTERN"; return $EXIT_ERR; fi
    fi

    # Getting all S/W users.
    local pdbusers=$($PDBEDIT -L)
    # Enumerating PDB users and sorting by S/W username.
    echo -e "$pdbusers" | $SORT -t ':' -k 3 | while read pdbuser
    do
	# Getting corresponding entity record via getentity function (no warnings).
	entity="$(getentity -lu `echo "$pdbuser" | $CUT -f1 -d ':'` 2>/dev/null)"
	# Applying filter (if any).
	if [ -n "$option" ]; then
	    case "$option" in
		# Filtering by Linux username.
		"lu"|"linuxuser")
		    linusername="$(echo -e "$entity" | $CUT -f2 -d ':')"
		    # Clear unmatching entity.
		    if [ -z "$(echo -e "$linusername" | $GREP -e "$pattern")" ]; then entity=""; fi
		;;
		# Filtering by Samba/Windows username (case-insensitive).
		"wu"|"sambauser"|"winuser")
		    winusername="$(echo -e "$entity" | $CUT -f4 -d ':')"
		    # Clear unmatching entity.
		    if [ -z "$(echo -e "$winusername" | $GREP -ie "$pattern")" ]; then entity=""; fi
		;;
	    esac
	fi
	# Printing entity(if not filtered above).
	if [ -n "$entity" ]; then echo -e "$entity"; fi
    done
    # Listing done.
    return
}

# smbm_listmemberships "S/W username"
# Listing all group memberships of a Samba/Windows user.
function smbm_listmemberships {

    # S/W username is mandatory and must be existing.
    local winusername="$1"; shift
    if [ -z "$winusername" ]; then log "$MSG_SMBM_ALL_NOWINUSER"; return $EXIT_ERR; fi
    local userentity=$(getentity -wu "$winusername" 2>/dev/null)
    if [ -z "$userentity" ]; then log "$MSG_SMBM_ALL_NOWINUSER"; return $EXIT_ERR; fi

    # Getting group SID list by net groupmap command.
    local groupids=$($GROUPMAP memberships $(echo -e "$userentity" | $CUT -f5 -d ':'))
    # Converting SID list into entity list.
    local entities=$(echo -e "$groupids" | while read groupid
                     do
                	getentity -id "$groupid" 2>/dev/null
                     done)
    # On success print results sorted by S/W groupname.
    if [ -n "$entities" ]; then
        echo -e "$entities" | $SORT -k4 -t ':'
    fi
    # Listing done.
    return
}

# smbm addgroup S/W_groupname [Linux_groupname]
# Creating and mapping a Samba/Windows group.
function smbm_addgroup {

    local domain=`$HOSTNAME | $AWK '{ print toupper($0) }'`
    local result

    # S/W groupname is mandatory and formally checked.
    local wingroupname="$1"; shift
    if [ -z "$wingroupname" ]; then log $MSG_SMBM_ALL_NOWINGROUP; return $EXIT_ERR; fi
    if [ ! -z "$(echo -e "$wingroupname" | $SED "s/$MASKWINGROUP//")" ]; then log $MSG_SMBM_ALL_NOWINGROUP; return $EXIT_ERR; fi
    # Giving up with existing S/W group.
    local mapentry=`getentity -wg "$wingroupname" 2>/dev/null`
    if [ -n "$mapentry" ]; then log "$MSG_SMBM_ALL_ENTITY_EXISTS $mapentry"; return $EXIT_ERR; fi

    # Get next available SID-RID and computes corresponding Linux GID.
    local localsid=$($NET getlocalsid | $SED "s/.*: //")
    local nextrid=`getnextrid`
    if [ -z "$nextrid" ]; then log "$MSG_SMBM_ALL_OUTOFRID"; return $EXIT_ERR;fi
    local lingid=$(( LINUIDBASE + nextrid ))
    # Making a corresponding Linux username.
    local lingroupname="$1"; shift
    # If unspecified, generates an unique RID mapped Linux username.
    if [ -z "$lingroupname" ]; then lingroupname="$LINGROUP$lingid"; fi
    # Formal check, giving up on failure.
    if [ ! -z "$(echo -e "$lingroupname" | $SED "s/$MASKLINGROUP//")" ]; then log $MSG_SMBM_ALL_NOLINGROUP; return $EXIT_ERR; fi

    # Check for existence of Linux group given/planned.
    local grpentry=`getlingroup "$lingroupname"`
    # Creating Linux group if necessary.
    if [ -z "$grpentry" ]; then
	$GROUPADD --gid $lingid "$lingroupname" 2>/dev/null; result=$?
	# On failure giving up.
	if [ ! $result -eq 0 ]; then
	    log "$MSG_SMBM_CONF_LINGROUP_FAILED $lingroupname!"
	    log "$MSG_SMBM_ALL_ERROR $result!"; return $result
	fi
    fi

    # Mapping S/W group to Linux group.
    $GROUPMAP add ntgroup="$wingroupname" sid="$localsid-$nextrid" type="l" \
              unixgroup=$lingroupname >/dev/null; result=$?
    # On failure giving up.
    if [ ! $result -eq 0 ]; then
	log "$MSG_SMBM_ALL_WINGROUPMAP_FAILED $wingroupname ($localsid-$nextrid) -> $lingroupname"
	return $result
    fi
    # Incrementing next available RID in S/W passdb backend.
    setnextrid "$(( nextrid+1 ))"; result=$?

    # Final check.
    local mapentry=`getentity -wg "$wingroupname" 2>/dev/null`; result=$?
    if [ $result -eq 0 ]; then
	log "$MSG_SMBM_ALL_ENTITY_CREATED $mapentry"
    else
	log "$MSG_SMBM_ALL_ENTITY_PROBLEM g:$lingroupname:$lingid:$wingroupname:$localsid-$nextrid!"
	log "$MSG_SMBM_ALL_ERROR $result!"; return $result
    fi

    return
}

# smbm_delgroup S/W_groupname
# Deleting a S/W group and corresponding Linux group.
function smbm_delgroup {

    local domain=`$HOSTNAME | $AWK '{ print toupper($0) }'`
    local message
    local result

    # S/W groupname is mandatory and must be existing.
    local wingroupname="$1"; shift
    if [ -z "$wingroupname" ]; then log "$MSG_SMBM_ALL_NOWINGROUP"; return $EXIT_ERR; fi
    local groupentity=$(getentity -wg "$wingroupname" 2>/dev/null)
    if [ -z "$groupentity" ]; then log "$MSG_SMBM_ALL_NOWINGROUP"; return $EXIT_ERR; fi
    # Log marker: start processing.
    log "$MSG_SMBM_ALL_ENTITY_DELETING $groupentity"

    # Emptying S/W group.
    local members=$(smbm_listmembers "$wingroupname")
    # Enumerating group members (if any) and deleting memberships.
    for member in $members
    do
	smbm_delmember "$(echo -e "$member" | $CUT -f4 -d ':')" "$wingroupname"
    done
    # S/W group is empty.

    # Removes group mapping via net groupmap call, destroying S/W entity.
    message=`$GROUPMAP delete ntgroup="$wingroupname" 2>&1`; result=$?
    if [ -n "$message" ]; then log "$message"; fi
    # Giving up on failure.
    if [ ! $result -eq 0 ]; then
	log "$MSG_SMBM_ALL_ENTITY_DELFAILED $groupentity"
	log "$MSG_SMBM_ALL_ERROR $result!"; return $result
    fi
    # Windows entity destroyed.

    # Removes corresponding file system ACLs in whole Samba area.
    message=`$SETFACL -R -x "g:$(echo -e "$groupentity" | $CUT -f2 -d ':')" "$SMBROOT" 2>&1`; result=$?
    if [ -n "$message" ]; then log "$message"; fi
    # Giving up on failure.
    if [ ! $result -eq 0 ]; then
	log "$MSG_SMBM_ALL_ENTITY_DELFAILED $groupentity"
	log "$MSG_SMBM_ALL_ERROR $result!"; return $result
    fi
    # Removes default ACLs also.
    message=`$SETFACL -R -d -x "g:$(echo -e "$groupentity" | $CUT -f2 -d ':')" "$SMBROOT" 2>&1`; result=$?
    if [ -n "$message" ]; then log "$message"; fi
    # Giving up on failure.
    if [ ! $result -eq 0 ]; then
	log "$MSG_SMBM_ALL_ENTITY_DELFAILED $groupentity"
	log "$MSG_SMBM_ALL_ERROR $result!"; return $result
    fi
    # Linux group has no file system rights in Samba area.

    # Removes Linux group itself.
    message=`$GROUPDEL "$(echo -e "$groupentity" | $CUT -f2 -d ':')" 2>&1`; result=$?
    if [ -n "$message" ]; then log "$message"; fi
    # Report success or failure.
    if [ $result -eq 0 ]; then
	log "$MSG_SMBM_ALL_ENTITY_DELETED $groupentity"
    else
	log "$MSG_SMBM_ALL_ENTITY_DELFAILED $groupentity"
	log "$MSG_SMBM_ALL_ERROR $result!"; return $result
    fi
    return
}

# smbm_modgroup S/W_groupname [ S/W_group_new_name ]
# Modifying (actually renaming) an existing Samba/Windows group.
function smbm_modgroup {

    local message
    local result=0

    # S/W groupname is mandatory and must be existing.
    local wingroupname="$1"; shift
    if [ -z "$wingroupname" ]; then log "$MSG_SMBM_ALL_NOWINGROUP"; return $EXIT_ERR; fi
    local groupentity=$(getentity -wg "$wingroupname" 2>/dev/null)
    if [ -z "$groupentity" ]; then log "$MSG_SMBM_ALL_NOWINGROUP"; return $EXIT_ERR; fi
    # Identified S/W group to modify.

    # New S/W groupname optional but must be unique and formally checked.
    local newgroupname="$1"; shift
    if [ -n "$wingroupname" ]; then
	# Formal check.
	if [ ! -z "$(echo -e "$newgroupname" | $SED "s/$MASKWINGROUP//")" ]; then 
	    log "$MSG_SMBM_ALL_BADWINGROUP $newgroupname"
	    return $EXIT_ERR
	fi
        # Giving up when it's another existing S/W group's name.
	local mapentity=`getentity -wg "$newgroupname" 2>/dev/null`
	if [ -n "$mapentity" -a ! "$mapentity" = "$groupentity" ]; then
	    log "$MSG_SMBM_ALL_ENTITY_EXISTS $mapentity"
	    return $EXIT_ERR
	fi
    fi
    # New groupname is valid (if any).

    # Renaming an empty group if requested.
    if [ -n "$newgroupname" ]; then
	# Giving up if group have members (TODO!).
	local members=$(smbm_listmembers "$wingroupname")
	if [ -n "$members" ]; then
	    log "$MSG_SMBM_MODGROUP_HASMEMBERS\n$members"
	    return $EXIT_ERR
	fi

	# Deleting old mapentry.
	message=$($GROUPMAP delete sid=$(echo -e "$groupentity" | $CUT -f5 -d ':'))
	result=$?
	# Logging message if any; ignoring errors here.
	if [ -n "$message" ]; then log "$message"; fi
	# Mapping S/W group to Linux group.
	message=$($GROUPMAP add ntgroup="$newgroupname" \
	                    sid=$(echo -e "$groupentity" | $CUT -f5 -d ':') \
	                    type="l" \
	                    unixgroup=$(echo -e "$groupentity" | $CUT -f2 -d ':'))
	result=$?
	# Reporting success or failure.
	if [ -n "$message" ]; then log "$message"; fi
	if [ $result -eq 0 ]; then
	    log "$MSG_SMBM_ALL_ENTITY_MODIFIED $groupentity -> " \
	        "$(getentity -wg "$newgroupname" 2>/dev/null)"
	else
	    log "$MSG_SMBM_ALL_WINGROUPMAP_FAILED " \
	        "$newgroupname ($(echo -e "$groupentity" | $CUT -f5 -d ':')) ->" \
		"$(echo -e "$groupentity" | $CUT -f2 -d ':')"
	    log "$MSG_SMBM_ALL_ERROR $result!"
	    # On failure giving up.
	    return $result
	fi
    fi
    # Renamed.

    return
}

# smbm_listgroup [ -lg|-wg regexp_pattern ]
# Listing entities with matching Linux OR Samba/Windows group names to pattern.
function smbm_listgroup {

    local mapentry
    local lingroupname
    local wingroupname

    # Getting optional option switch.
    local option=`echo $1 | $SED -n "s/^-\+//p"`
    # If option given a pattern is mandatory.
    local pattern=""
    if [ -n "$option" ]; then
	shift; pattern="$1"
	# Giving up on missing pattern.
	if [ -z "$pattern" ]; then log "$MSG_SMBM_LIST_NOPATTERN"; return $EXIT_ERR; fi
    fi

    # Getting all S/W groups.
    local groupmap="$($GROUPMAP list)"
    # Enumerating groups and sorting by S/W group name.
    echo -e "$groupmap" | $SORT | while read mapentry
    do
	# Getting corresponding entity record via getentity function (no warnings).
	entity=`getentity -wg $(echo "$mapentry" | $SED "s/^\(.*\) \([^ ]*\)$/\2/") 2>/dev/null`
	# Applying filter (if any).
	if [ -n "$option" ]; then
	    case "$option" in
		# Filtering by Linux group name.
		"lg"|"linuxgroup")
		    lingroupname="$(echo -e "$entity" | $CUT -f2 -d ':')"
		    # Clear unmatching entity.
		    if [ -z "$(echo -e "$lingroupname" | $GREP -e "$pattern")" ]; then entity=""; fi
		;;
		# Filtering by Samba/Windows group name (case-insensitive).
		"wg"|"sambagroup"|"wingroup")
		    wingroupname="$(echo -e "$entity" | $CUT -f4 -d ':')"
		    # Clear unmatching entity.
		    if [ -z "$(echo -e "$wingroupname" | $GREP -ie "$pattern")" ]; then entity=""; fi
		;;
	    esac
	fi
	# Printing entity (if not filtered above).
	if [ -n "$entity" ]; then echo -e "$entity"; fi
    done
    # Listing done.
    return
}

# smbm_addmember S/W_username S/W_groupname
# Adding a S/W user to a S/W group with underlying Linux membership.
function smbm_addmember {

    local result

    # S/W username is mandatory and must be existing.
    local winusername="$1"; shift
    if [ -z "$winusername" ]; then log "$MSG_SMBM_ALL_NOWINUSER"; return $EXIT_ERR; fi
    local userentity=$(getentity -wu "$winusername" 2>/dev/null)
    if [ -z "$userentity" ]; then log "$MSG_SMBM_ALL_NOWINUSER"; return $EXIT_ERR; fi

    # S/W group name is mandatory and must be existing.
    local wingroupname="$1"; shift
    if [ -z "$wingroupname" ]; then log "$MSG_SMBM_ALL_NOWINGROUP"; return $EXIT_ERR; fi
    local groupentity=$(getentity -wg "$wingroupname" 2>/dev/null)
    if [ -z "$groupentity" ]; then log "$MSG_SMBM_ALL_NOWINGROUP"; return $EXIT_ERR; fi

    # Establishing Linux membership via adduser call.
    $ADDUSER "$(echo -e "$userentity" | $CUT -f2 -d ':')" \
             "$(echo -e "$groupentity" | $CUT -f2 -d ':')" >/dev/null 2>&1;
    result=$?
    # On failure giving up vith log entry.
    if [ ! $result -eq 0 ]; then
	log "$MSG_SMBM_ALL_MEMBERSHIP_FAILED $userentity -> $groupentity"
	log "$MSG_SMBM_ALL_ERROR $result!"
	return $result
    fi

    # Establishing S/W membership via net groupmap call.
    $GROUPMAP addmem \
                  "$(echo -e "$groupentity" | $CUT -f5 -d ':')" \
                  "$(echo -e "$userentity" | $CUT -f5 -d ':')" >/dev/null 2>&1; \
                  result=$?
    # On failure giving up vith log entry.
    if [ ! $result -eq 0 ]; then
	log "$MSG_SMBM_ALL_MEMBERSHIP_FAILED $userentity -> $groupentity"
	log "$MSG_SMBM_ALL_ERROR $result!"
	return $result
    fi

    # Reporting success and finish.
    log "$MSG_SMBM_ALL_MEMBERSHIP_CREATED $userentity -> $groupentity"
    return $result
}

# smbm_delmember S/W_username S/W_groupname
# Deleting a S/W user from a S/W group with underlying Linux membership also.
function smbm_delmember {

    local result

    # S/W username is mandatory and must be existing.
    local winusername="$1"; shift
    if [ -z "$winusername" ]; then log "$MSG_SMBM_ALL_NOWINUSER"; return $EXIT_ERR; fi
    local userentity=$(getentity -wu "$winusername" 2>/dev/null)
    if [ -z "$userentity" ]; then log "$MSG_SMBM_ALL_NOWINUSER"; return $EXIT_ERR; fi

    # S/W group name is mandatory and must be existing.
    local wingroupname="$1"; shift
    if [ -z "$wingroupname" ]; then log "$MSG_SMBM_ALL_NOWINGROUP"; return $EXIT_ERR; fi
    local groupentity=$(getentity -wg "$wingroupname" 2>/dev/null)
    if [ -z "$groupentity" ]; then log "$MSG_SMBM_ALL_NOWINGROUP"; return $EXIT_ERR; fi

    # Deleting Linux membership via deluser call.
    $DELUSER "$(echo -e "$userentity" | $CUT -f2 -d ':')" \
             "$(echo -e "$groupentity" | $CUT -f2 -d ':')" >/dev/null 2>&1;
    result=$?
    # On failure giving up vith log entry.
    if [ ! $result -eq 0 ]; then
	log "$MSG_SMBM_ALL_MEMBERSHIP_DELFAILED $userentity -> $groupentity"
	log "$MSG_SMBM_ALL_ERROR $result!"
	return $result
    fi

    # Deleting S/W membership via net groupmap call.
    $GROUPMAP delmem \
                  "$(echo -e "$groupentity" | $CUT -f5 -d ':')" \
                  "$(echo -e "$userentity" | $CUT -f5 -d ':')" >/dev/null 2>&1; \
                  result=$?
    # On failure giving up vith log entry.
    if [ ! $result -eq 0 ]; then
	log "$MSG_SMBM_ALL_MEMBERSHIP_DELFAILED $userentity -> $groupentity"
	log "$MSG_SMBM_ALL_ERROR $result!"
	return $result
    fi

    # Reporting success and finish.
    log "$MSG_SMBM_ALL_MEMBERSHIP_DELETED $userentity -X- $groupentity"
    return
}

# smbm_listmembers S/W_groupname
# Listing member entities of a S/W group.
function smbm_listmembers {

    # S/W group name is mandatory and must be existing.
    local wingroupname="$1"; shift
    if [ -z "$wingroupname" ]; then log "$MSG_SMBM_ALL_NOWINGROUP"; return $EXIT_ERR; fi
    local groupentity=$(getentity -wg "$wingroupname" 2>/dev/null)
    if [ -z "$groupentity" ]; then log "$MSG_SMBM_ALL_NOWINGROUP"; return $EXIT_ERR; fi

    # Getting member SID list by net groupmap command.
    local memberids=$($GROUPMAP listmem $(echo -e "$groupentity" | $CUT -f5 -d ':'))
    # Converting SID list into entity list.
    local entities=$(echo -e "$memberids" | while read memberid
                     do
                	getentity -id "$memberid" 2>/dev/null
                     done)
    # On success print results sorted by S/W username.
    if [ -n "$entities" ]; then
        echo -e "$entities" | $SORT -k4 -t ':'
    fi
    # Listing done.
    return
}


# smbm_configure
# Graceful setup or reconfiguration of smbm environment.
function smbm_configure {

# Checking components and circumstances.
    if [ ! $UID -eq 0 ]; then echo -e "$MSG_SMBM_ALL_ROOTNEED" >&2; return $EXIT_ERR; fi

# Logs handling: smbm log creation (rotated via dedicated cronjob).
# Must be the 1st step to enable logging configuration process itself.
    if [ ! -f "$SMBMLOGFILE" ]; then
        touch "$SMBMLOGFILE"; chown root:root "$SMBMLOGFILE"; chmod 640 "$SMBMLOGFILE"
        log "$MSG_SMBM_CONF_LOGFILE_CREATED" "$SMBMLOGFILE"
    fi

# Creating timestamp to name backup files and logfile.
    local timestamp="$(date '+%Y%m%d%H%M%S')"

# Stop Samba during configuration.
    log $($SMBDAEMONS stop 2>/dev/null)

# Creating Samba owner Linux system user if isn't exist.
    # Dirty trick to check existence.
    local result=$(su $SMBUSER -s /bin/bash -c "echo" 2>/dev/null; echo $?)
    if [ ! $result -eq 0 ]; then
	# Adding a Linux system user with no shell and no valid password.
	$USERADD --system -c "$SMBADMIN" -g "$SMBGROUP" -d "$SMBHOME" -s "/bin/false" "$SMBUSER"; result=$?
	if [ $result -eq 0 ]; then
	    log "$MSG_SMBM_CONF_LINUSER_CREATED $SMBUSER."
	else
	    log "$MSG_SMBM_CONF_LINUSER_FAILED $SMBUSER!"; return $EXIT_ERR
	fi
    else
	log "$MSG_SMBM_CONF_LINUSER_EXISTS $SMBUSER."
    fi
    # Checking and creating home directory if absent.
    if [ ! -d "$SMBHOME" ]; then
	log $(mkdir --verbose -m 2771 "$SMBHOME" 2>&1)
	chown $SMBUSER:$SMBGROUP "$SMBHOME"
    fi
    # Checking and creating .ssh folder and trusted keys file if absent.
    if [ ! -d "$SMBHOME/$SSHKEYSDIR" ]; then
	log $(mkdir --verbose -m 2700 "$SMBHOME/$SSHKEYSDIR" 2>&1)
	chown $SMBUSER:$SMBGROUP "$SMBHOME/$SSHKEYSDIR"
    fi
    if [ ! -f "$SMBHOME/$SSHKEYSDIR/$SSHKEYSFILE" ]; then
	log $(touch "$SMBHOME/$SSHKEYSDIR/$SSHKEYSFILE" 2>&1)
	chown $SMBUSER:$SMBGROUP "$SMBHOME/$SSHKEYSDIR/$SSHKEYSFILE"
	chmod 400 "$SMBHOME/$SSHKEYSDIR/$SSHKEYSFILE"
	chattr +i "$SMBHOME/$SSHKEYSDIR/$SSHKEYSFILE"
    fi
    # Configured Samba owner.

# Creating Samba storage.
    if [ ! -d "$SMBROOT" ]; then
	log $(mkdir --verbose -m 2771 "$SMBROOT" 2>&1)
	chown $SMBUSER:$SMBGROUP "$SMBROOT"
    fi
    for folder in "$SMBDIR_HOMES" "$SMBDIR_PROFILES" "$SMBDIR_SERVICES" "$SMBDIR_SHARE"
    do
	if [ ! -d "$SMBROOT/$folder" ]; then
	    log $(mkdir --verbose -m 2771 "$SMBROOT/$folder" 2>&1) 
	    chown $SMBUSER:$SMBGROUP "$SMBROOT/$folder"
	fi
    done
    # Configured Samba storage.

# Creating basic configuration, but once only.
    # Need to create if absent.
    if [ ! -f "$SMBCONFDIR/$SMBCONFFILE" ]; then
	# Creating from template.
        export TIMESTAMP=$(date '+%Y-%m-%d %H:%M:%S')
	export localdomain=`$HOSTNAME | $AWK '{ print toupper($0) }'`
	export localip=`echo $(LANG=C $IFCONFIG | $GREP 'inet addr' | $SED "s/ *inet addr:\([0-9\.]*\).*$/\1/")`
	export SMBCONFDIR SMBDIR_SHARE SMBGROUP SMBMAPFILE SMBROOT SMBUSER
        cat $SMBTEMPLATES/$SMBCONFFILE | $ENVSUBST > $SMBCONFDIR/$SMBCONFFILE
        chmod 640 $SMBCONFDIR/$SMBCONFFILE
	log $MSG_SMBM_CONF_CONFFILE_CREATED $SMBCONFDIR/$SMBCONFFILE.
    else
	# Config file exists, but maybe it is the maintaner's.
	# In this case there is no backup files.
	if [ -z "$(ls $SMBCONFDIR/$SMBCONFFILE*.bak 2>/dev/null)" ]; then
	    # Making backup.
	    log $(mv --verbose "$SMBCONFDIR/$SMBCONFFILE" "$SMBCONFDIR/$SMBCONFFILE.$timestamp.bak" 2>&1)
	    # Creating from template.
    	    export TIMESTAMP=$(date '+%Y-%m-%d %H:%M:%S')
	    export localdomain=`$HOSTNAME | $AWK '{ print toupper($0) }'`
	    export localip=`echo $(LANG=C $IFCONFIG | $GREP 'inet addr' | $SED "s/ *inet addr:\([0-9\.]*\).*$/\1/")`
	    export SMBCONFDIR SMBDIR_SHARE SMBGROUP SMBMAPFILE SMBROOT SMBUSER
    	    cat $SMBTEMPLATES/$SMBCONFFILE | $ENVSUBST > $SMBCONFDIR/$SMBCONFFILE
    	    chmod 640 $SMBCONFDIR/$SMBCONFFILE
	    log $MSG_SMBM_CONF_CONFFILE_CREATED $SMBCONFDIR/$SMBCONFFILE.
	# If backup exists Samba maybe already configured;
	# isn't safe to overwrite configuration.
	else
	    log $MSG_SMBM_CONF_CONFFILE_EXISTS
	fi
    fi
    # Checking configuration (always).
    result=$($TESTPARM -s 1>/dev/null 2>&1; echo $?)
    if [ ! $result -eq 0 ]; then
	log $MSG_SMBM_CONF_CONFFILE_FAILED $SMBCONFDIR/$SMBCONFFILE!
	return $EXIT_ERR
    fi
    # Verified existing or just created basic configuration.

# Creating user map file if absent.
    if [ ! -f $SMBCONFDIR/$SMBMAPFILE  ]; then
    	export TIMESTAMP=$(date '+%Y-%m-%d %H:%M:%S')
	export SMBUSER SMBADMIN
	cat $SMBTEMPLATES/$SMBMAPFILE | $ENVSUBST > $SMBCONFDIR/$SMBMAPFILE
    	chmod 640 $SMBCONFDIR/$SMBMAPFILE
	log $MSG_SMBM_CONF_MAPFILE_CREATED $SMBCONFDIR/$SMBMAPFILE
    else
	log $MSG_SMBM_CONF_MAPFILE_EXISTS
    fi
    # Existing user map file.

# User database revision.
    # Checking Samba/Windows Administrator account.
    result=$($PDBEDIT -L $SMBUSER 1>/dev/null 2>&1; echo $?)
    if [ $result -eq 0 ]; then
	log $MSG_SMBM_CONF_SMBUSER_EXISTS $SMBUSER.
    else
	# Creating Samba/Windows local administrator with RID=500 and generated password.
	local localsid=$($NET getlocalsid | $SED "s/.*: //")
	local adminpass=$($PWGEN -n -c -s 12 1)
	result=$(echo -e "$adminpass\n$adminpass" | \
	         $PDBEDIT -a -t -u $SMBUSER -U "$localsid-500" -f "$SMBADMIN" \
	                  -h "$SMBROOT/$SMBDIR_HOMES/$SMBADMIN" -p "$SMBROOT/$SMBDIR_PROFILES/$SMBADMIN" -D $SMBHOMEDRIVE 1>/dev/null; echo $?)
	# Results and password are logged.
	if [ $result -eq 0 ]; then
	    log "$MSG_SMBM_CONF_SMBUSER_CREATED $SMBUSER ($localsid-500) $adminpass"
	else
	    log "$MSG_SMBM_CONF_SMBUSER_FAILED $SMBUSER ($localsid-500) $adminpass"
	    return $EXIT_ERR
	fi
    fi
    # Existing Samba/Windows account for administration.

    # Checking builtin administrators group mapping.
    if [ -z "$($NET groupmap list | $GREP "$SMBADMINGROUPSID")" ]; then
	result=$($NET groupmap add ntgroup="$SMBADMINGROUP" sid="$SMBADMINGROUPSID" type="b" unixgroup=$SMBGROUP >/dev/null; echo $?)
	if [ $result -eq 0 ]; then
	    log $MSG_SMBM_CONF_GROUPMAP_CREATED
	else
	    log $MSG_SMBM_CONF_GROUPMAP_FAILED
	    return $EXIT_ERR
	fi
    else
	log "$MSG_SMBM_CONF_GROUPMAP_EXISTS"
    fi
    # Builtin administrators group is mapped.

    # Checking Samba administrator's membership.
    local localsid=$($NET getlocalsid | $SED "s/.*: //")
    if [ -z "$($NET groupmap listmem $SMBADMINGROUPSID | $GREP "$localsid-500")" ]; then
	result=$($NET groupmap addmem "$SMBADMINGROUPSID" "$localsid-500" 2>/dev/null; echo $?)
	if [ $result -eq 0 ]; then
	    log "$MSG_SMBM_CONF_ADMINMEMBER_CREATED ($localsid-500 -> $SMBADMINGROUPSID)."
	else
	    log "$MSG_SMBM_CONF_ADMINMEMBER_FAILED ($localsid-500 -> $SMBADMINGROUPSID)!"
	    return $EXIT_ERR
	fi
    else
	log "$MSG_SMBM_CONF_ADMINMEMBER_EXISTS"
    fi
    # Samba/Windows Administrator is a member of Builtin Admins group.

    # Looking for unnecessary Samba accounts.
    # Samba users with Linux UID 1000-1999 (maybe created by Debian setup)
    # considered as unnecessary accounts.
    local users1000_1999=$($PDBEDIT -L | $GREP ':1[[:digit:]]\{3\}:')
    if [ ! -z "$users1000_1999" ]; then
	for user in $users1000_1999
	do
	    log "$MSG_SMBM_CONF_EXTRAUSER $user"
	done
    fi
    # Warnings are sent about (maybe) unnecessary Samba/Windows accounts.

# Finally go on with Samba!
    log $($SMBDAEMONS start)
    return
}

# Helper function: retrieve a complete entity record.
# Usage: getentity -lu Linux_username
#        getentity -lg Linux_groupname
#        getentity -wu "Samba/Windows username"
#        getentity -wg "Samba/Windows groupname"
#        getentity -id Samba/Windows_SID-RID
# Returns with a colon-delimited entity record or empty string.
# Format: type:Linux_name:UID/GID:Windows name:SID-RID
function getentity {

    # Getting mandatory option switch.
    local option=`echo $1 | $SED -n "s/^-\+//p"`; if [ ! -z "$option" ]; then shift; fi
    local result=0

    case "$option" in

	# Getting entry by Linux username.
	"lu"|"linuxuser")
	    # Empty username makes an immediate error.
	    if [ -z "$1" ]; then 
		result=1
		log "$MSG_SMBM_ALL_ERROR $result!"
		return $result
	    fi
	    # Linux authentication database (/etc/passwd) entry.
	    local pwdentry=$(getlinuser "$1" 2>/dev/null)
	    # Missing pwdentry makes a negative answer.
	    if [ -z "$pwdentry" ]; then
		result=$(( result + 3 ))
		log "$MSG_SMBM_ALL_ERROR $result!"
		return $result
	    fi
	    # Samba/Windows authentication database (passdb.tdb) entry.
	    local pdbentry=$($PDBEDIT -L "$1" 2>/dev/null)
	    # Consistency check:
	    # Missing pdbentry is unacceptable.
	    if [ -z "$pdbentry" ]; then
		result=$(( result + 5 ))
		log "$MSG_SMBM_ALL_ERROR $result!"
		return $result
	    fi
	    # Warning about inconsistent Gecos and PDB Comment fields.
	    if [ ! "$pwdentry" == "$pdbentry" ]; then result=$(( result + 8 )); fi
	    # Retrieve SID-RID.
	    local winid=`$PDBEDIT -Lv "$1" | $GREP 'User SID:' | \
	                 $SED "s/^\(.*\): *\(S-.*\)$/\2/"`
	    # Check for username mapping.
	    local mapentry=$(getwinuser "$1" 2>/dev/null)
	    # Missing alias replaced with Linux username.
	    if [ -z "$mapentry" ]; then mapentry="$1:$1"; fi
	    # Consistency check:
	    # Warning about inconsistent alias and Gecos field.
	    if [ ! "$(echo -e "$mapentry" | $CUT -f2 -d ':')" == \
	            "$(echo -e "$pdbentry" | $CUT -f3 -d ':')" ]; then
		result=$(( result + 16 ))
	    fi
	    # Warning about inconsistent alias and PDB Comment field.
	    if [ ! "$(echo -e "$mapentry" | $CUT -f2 -d ':')" == \
	            "$(echo -e "$pwdentry" | $CUT -f3 -d ':')" ]; then
		result=$(( result + 32 ))
	    fi
	    # Assembling entry.
	    echo -ne "u:"					# Type = user
	    echo -ne "$(echo "$pwdentry" | $CUT -f1,2 -d ':'):"	# Linux data
	    echo -ne "$(echo "$mapentry" | $CUT -f2 -d ':'):"	# S/W username
	    echo -ne "$winid"					# S/W SID-RID
	    echo -ne "\n"					# Closing entry

	    # Result codes are:
	    # bit 1 ( 1): error (otherwise warning only)
	    # bit 2 ( 2): no such Linux user (makes error)
	    # bit 3 ( 4): missing PDB entry (makes error)
	    # bit 4 ( 8): inconsistent Gecos and PDB comment fields
	    # bit 5 (16): inconsistent alias and Gecos field
	    # bit 6 (32): inconsistent alias and PDB comment field
	;;

	# Getting entry by Linux groupname.
	"lg"|"linuxgroup")
	    # Empty groupname makes an immediate error.
	    if [ -z "$1" ]; then 
		result=1
		log "$MSG_SMBM_ALL_ERROR $result!"
		return $result
	    fi
	    # Linux group database (/etc/group) entry.
	    local grpentry=$(getlingroup "$1" 2>/dev/null)
	    # Missing grpentry makes a negative answer.
	    if [ -z "$grpentry" ]; then
		result=$(( result + 3 ))
		log "$MSG_SMBM_ALL_ERROR $result!"
		return $result
	    fi
	    # Samba/Windows groupmap (group.mapping.ldb) entry.
	    local gdbentry=$(getwingroup "$1")
	    # Missing gdbentry is unacceptable.
	    if [ -z "$gdbentry" ]; then
		result=$(( result + 5 ))
		log "$MSG_SMBM_ALL_ERROR $result!"
		return $result
	    fi
	    # Assembling entry.
	    echo -ne "g:"					# Type = group
	    echo -ne "$(echo "$grpentry" | $CUT -f1,2 -d ':'):"	# Linux data
	    echo -ne "$(echo "$gdbentry" | $CUT -f2,3 -d ':')"	# S/W data
	    echo -ne "\n"					# Closing entry

	    # Result codes are:
	    # bit 1 ( 1): error (otherwise warning only)
	    # bit 2 ( 2): no such Linux group (makes error)
	    # bit 3 ( 4): missing groupmap entry (makes error)
	;;

	# Getting entry by Windows username.
	"wu"|"sambauser"|"windowsuser")
	    # Empty username makes an immediate error.
	    if [ -z "$1" ]; then
		result=1
		log "$MSG_SMBM_ALL_ERROR $result!"
		return $result
	    fi
	    # Check for username mapping.
	    local mapentry=$(getwinuser "$1" 2>/dev/null)
	    # Missing alias replaced with username given.
	    if [ -z "$mapentry" ]; then mapentry="$1:$1"; fi
	    # Linux username derived from mapentry.
	    local linusername=$(echo "$mapentry" | $CUT -f1 -d ':')
	    # Samba/Windows authentication database (passdb.tdb) entry.
	    local pdbentry=$($PDBEDIT -L "$linusername" 2>/dev/null)
	    # Missing pdbentry is unacceptable.
	    if [ -z "$pdbentry" ]; then
		result=$(( result + 5 ))
		log "$MSG_SMBM_ALL_ERROR $result!"
		return $result
	    fi
	    # Retrieve SID-RID.
	    local winid=`$PDBEDIT -Lv "$linusername" | $GREP 'User SID:' | \
	                 $SED "s/^\(.*\): *\(S-.*\)$/\2/"`
	    # Consistency check:
	    # must be a corresponding Linux pwdentry.
	    local pwdentry=$(getlinuser "$linusername" 2>/dev/null)
	    # Missing pwdentry makes a negative answer.
	    if [ -z "$pwdentry" ]; then
		result=$(( result + 3 ))
		log "$MSG_SMBM_ALL_ERROR $result!"
		return $result
	    fi
	    # Warning about inconsistent Gecos and PDB Comment fields.
	    if [ ! "$pwdentry" == "$pdbentry" ]; then result=$(( result + 8 )); fi
	    # Warning about inconsistent alias and Gecos field.
	    if [ ! "$(echo -e "$mapentry" | $CUT -f2 -d ':')" == \
	            "$(echo -e "pdbentry" | $CUT -f3 -d ':')" ]; then
		result=$(( result + 16 ))
	    fi
	    # Warning about inconsistent alias and PDB Comment field.
	    if [ ! "$(echo -e "$mapentry" | $CUT -f2 -d ':')" == \
	            "$(echo -e "pwdentry" | $CUT -f3 -d ':')" ]; then
		result=$(( result + 32 ))
	    fi
	    # Assembling entry.
	    echo -ne "u:"					# Type = user
	    echo -ne "$(echo "$pwdentry" | $CUT -f1,2 -d ':'):"	# Linux data
	    echo -ne "$(echo "$mapentry" | $CUT -f2 -d ':'):"	# S/W username
	    echo -ne "$winid"					# S/W SID-RID
	    echo -ne "\n"					# Closing entry

	    # Result codes are:
	    # bit 1 ( 1): error (otherwise warning only)
	    # bit 2 ( 2): no such Linux user (makes error)
	    # bit 3 ( 4): missing PDB entry (makes error)
	    # bit 4 ( 8): inconsistent Gecos and PDB comment fields
	    # bit 5 (16): inconsistent alias and Gecos field
	    # bit 6 (32): inconsistent alias and PDB comment field
	;;

	# Getting entry by Samba/Windows groupname.
	"wg"|"sambagroup"|"windowsgroup")
	    # Empty groupname makes an immediate error.
	    if [ -z "$1" ]; then 
		result=1
		log "$MSG_SMBM_ALL_ERROR $result!"
		return $result
	    fi
	    # Samba/Windows groupmap (group.mapping.ldb) entry.
	    local gdbentry=$(getwingroup "$1")
	    # Missing gdbentry makes a negative answer.
	    if [ -z "$gdbentry" ]; then
		result=$(( result + 5 ))
		log "$MSG_SMBM_ALL_ERROR $result!"
		return $result
	    fi
	    local lingroupname=$(echo "$gdbentry" | $CUT -f1 -d ':')
	    # Consistency check:
	    # must be a corresponding Linux group database (/etc/group) entry.
	    local grpentry=$(getlingroup "$lingroupname" 2>/dev/null)
	    # Missing grpentry makes a negative answer.
	    if [ -z "$grpentry" ]; then
		result=$(( result + 3 ))
		log "$MSG_SMBM_ALL_ERROR $result!"
		return $result
	    fi
	    # Assembling entry.
	    echo -ne "g:"					# Type = group
	    echo -ne "$(echo "$grpentry" | $CUT -f1,2 -d ':'):"	# Linux data
	    echo -ne "$(echo "$gdbentry" | $CUT -f2,3 -d ':')"	# S/W data
	    echo -ne "\n"					# Closing entry

	    # Result codes are:
	    # bit 1 ( 1): error (otherwise warning only)
	    # bit 2 ( 2): no such Linux group (makes error)
	    # bit 3 ( 4): missing groupmap entry (makes error)
	;;

	# Getting entry by Samba/Windows SID-RID.
	"id"|"rid"|"sid")
	    # Empty SID-RID makes an immediate error.
	    if [ -z "$1" ]; then 
		result=1
		log "$MSG_SMBM_ALL_ERROR $result!"
		return $result
	    fi
	    # Looking for group SID-RID by smbm call.
	    local entry=$(smbm_listgroup | $GREP -e ":$1$")
	    # On failure looking for user SID-RID by other smbm call.
	    if [ -z "$entry" ]; then
		entry=$(smbm_listuser | $GREP -e ":$1$")
	    fi
	    # On success printing entry.
	    if [ -n "$entry" ]; then echo -e "$entry"; fi
	;;

	*)
	    result=1
	    log "$MSG_SMBM_ALL_ERROR $result!"
	    return $result
	;;
    esac

    # Logs any unusual.
    if [ $result -ne 0 ]; then log "$MSG_SMBM_ALL_WARNING $result!"; fi
    return $result
}

# Helper function: find a Linux username in Linux authentication database.
# Usage: getlinuser case_sensitive_Linux_username
# Returns with a Linux_username:UID:Gecos comment entry or with an empty string.
function getlinuser {

    local linusername="$1"; shift

    # Search via getent.
    local lindbline=$($GETENT passwd $linusername)
    # Returns with username and Gecos fields only.
    if [ -n "$lindbline" ]; then
	echo -e "$lindbline" | $CUT -f1,3,5 -d ':'
    fi

    return
}

# Helper function: find a Linux groupname in Linux group datbase.
# Usage: getlingroup case_sensitive_Linux_groupname
# Returns with a Linux_group:GID entry or with an empty string.
function getlingroup {

    local lingroupname="$1"; shift

    # Search via getent.
    local lindbline=$($GETENT group "$lingroupname")
    # Returns with username and Gecos fields only.
    if [ -n "$lindbline" ]; then
	echo -e "$lindbline" | $CUT -f1,3 -d ':'
    fi

    return
}

# Helper function: find a user in Samba map (alias) file.
# Usage: getwinuser case insensitive S/W OR case sensitive Linux username
# Returns with a Linux username:S/W username entry or with an empty string.
function getwinuser {

    local username="$1"; shift

    local entry=""
    local linusername=""
    local winusername=""

    # Parsing Samba map (alias) file for case-insensitive Samba/Windows username.
    local aliasline=`cat $SMBCONFDIR/$SMBMAPFILE | $GREP -Ei "\"$username\"" `
    if [ -n "$aliasline" ]; then
	# Getting S/W username exactly from alias file.
	winusername=`echo -ne "$aliasline" | $SED "s/^\(.*\) = \"\(.*\)\".*$/\2/"`
	# Getting corresponding Linux user.
	linusername=`echo -ne "$aliasline" | $SED "s/^\(.*\) = .*$/\1/"`
	entry="$linusername:$winusername"
    fi

    # If isn't found, parsing for a case-sensitive Linux username.
    aliasline=`cat $SMBCONFDIR/$SMBMAPFILE | $GREP -E "^$username ="`
    if [ -n "$aliasline" ]; then
	linusername="$username"
	# Getting S/W username from alias file.
	winusername=`echo -ne "$aliasline" | $SED "s/^\(.*\) = \"\(.*\)\".*$/\2/"`
    fi

    # Show results (if any).
    if [ -n "$linusername" -a -n "$winusername" ]; then
	echo -e "$linusername:$winusername"
    fi

    return
}

# Helper function: find a group in Samba groupmap database.
# Usage: getwingroup case insensitive S/W OR case sensitive Linux groupname
# Returns with a Linux groupname:S/W groupname:SID-RID entry or with an empty string.
function getwingroup {

    local groupname="$1"; shift
    if [ -z "$groupname" ]; then return; fi

    # Parsing Samba groupmap for case-insensitive Samba/Windows groupname.
    local mapline=`$GROUPMAP list | $GREP -Ei "^$groupname " `
    # On failure looks again for case-sensitive Linux groupname.
    if [ -z "$mapline" ]; then
	local mapline=`$GROUPMAP list | $GREP -E "\-> $groupname$" `
    fi
    if [ -n "$mapline" ]; then
	# Getting S/W group name exactly from groupmap.
	local wingroupname=`echo -ne "$mapline" | $SED "s/^\(.*\) (.*$/\1/"`
	# Getting SID-RID from groupmap.
	local winid=`echo -ne "$mapline" | $SED "s/^.* (\(.*\)) .*$/\1/"`
	# Getting corresponding Linux group from groupmap.
 	local lingroupname=`echo -ne "$mapline" | $SED "s/^\(.*\) -> \(.*\)$/\2/"`
	# Assembling entry.
	echo -e "$lingroupname:$wingroupname:$winid"
    fi
}

# Helper function: handles Samba/Wndows -> Linux user mapping file.
# Usage: usermap "linusername:winusername" (adds or changes corresponding Linux entry)
#        usermap "linusername:" (deletes corresponding Linux entry)
function usermap {

    # Linux username is mandatory.
    local entry="$1"; shift
    if [ -z "$entry" ]; then return $EXIT_ERR; fi
    local linusername=$(echo -e "$entry" | $CUT -f1 -d ':')
    if [ -z "$linusername" ]; then return $EXIT_ERR; fi
    local winusername=$(echo -e "$entry" | $CUT -f2 -d ':')

    # Parsing Samba map file looking for case-sensitive Linux username.
    local aliasline=`cat $SMBCONFDIR/$SMBMAPFILE | $GREP -E "^$linusername ="`
    # If found we need to change (or delete) it.
    if [ -n "$aliasline" ]; then
	# Compiling new mapfile in a temporary file with unique filename.
	local tmpfile="$SMBCONFDIR/$SMBMAPFILE.$(date '+%N').tmp"
	# Aliasline removed.
	cat "$SMBCONFDIR/$SMBMAPFILE" | $GREP -Ev "^$linusername =" >>"$tmpfile"
	# If winusername given, new alias appended.
	if [ -n "$winusername" ]; then
	    echo -e "$linusername = \"$winusername\"" >> "$tmpfile"
	fi
	# Updates configfile.
	mv "$tmpfile" "$SMBCONFDIR/$SMBMAPFILE"
    # If isn't found and winusername given we need simply to append the new entry.
    else
	if [ -n "$winusername" ]; then
	    echo -e "$linusername = \"$winusername\"" >> "$SMBCONFDIR/$SMBMAPFILE"
	fi
    fi
    return
}

# Helper function: returns next available RID according to Samba/Windows backend.
function getnextrid {

    local nextrid
    # Available only for TDB backend so far.
    nextrid=`$PASSDB show "NEXT_RID\0" | \
	    $GREP '[000]' | \
	    $SED "s/^\[000\] \([[:xdigit:]]\{2\}\) \([[:xdigit:]]\{2\}\) \([[:xdigit:]]\{2\}\) \([[:xdigit:]]\{2\}\).*$/\4\3\2\1/"`
    nextrid=$(echo " ibase=16; $nextrid" | $BC)

    # Never goes over SMBRIDMAX value.
    if [ $nextrid -le $SMBRIDMAX ]; then
	echo -ne "$nextrid"; return
    else 
	return $EXIT_ERR
    fi
}

# Helper function: sets next available RID to parameter's (positive decimal) value.
function setnextrid {

    local nextrid="$1"; shift
    if [ -z "$nextrid" ]; then return $EXIT_ERR; fi

    # Available only for TDB backend so far.
    # 4 bytes hexadecimal conversion and endian change.
    nextrid=$(echo 00000000$(echo "ibase=10; obase=16; $nextrid" | $BC))
    # Emergency brake on failed hex conversion.
    if [ "$nextrid" = "00000000" ]; then return $EXIT_ERR; fi
    nextrid=`echo "$nextrid" | $SED "s/\(.*\)\(.\{2\}\)\(.\{2\}\)\(.\{2\}\)\(.\{2\}\)$/ \5 \4 \3 \2/"`
    # Backslash markers for tdbtool.
    nextrid=`echo "$nextrid" | $SED "s/ /\\\\\/g"`

    # Storing silently but reporting success or failure.
    $PASSDB store "NEXT_RID\0" "$nextrid" >/dev/null 2>&1; result=$?
    return $result
}

# Helper function: generates a keypair for tunnel authentication.
# Usage: keygen [ pathname [ passphrase [ comment ]]]
function keygen {

    # Optional pathname for keyfiles (defaults to ./keyfile).
    local keyfile=${1:-"keyfile"}; shift
    # Optional passphrase for keyfiles (defaults to random).
    local passphrase=${1:-$($MAKEPASSWD)}; shift
    # Optional comment (maybe empty).
    local comment="$1"; shift

    # OpenSSL/PuTTYGen used if available.
    if [ -x "$OPENSSL" -a -x "$PUTTYGEN" ]; then
	# Generating 2048 bits RSA2 key with DES-EDE3-CBC cipher.
	$OPENSSL genrsa -des3 -passout pass:"$passphrase" -out "$keyfile" 2048 >/dev/null 2>&1; result=$?
	if [ $result -ne 0 ]; then log $MSG_SMBM_KEYGEN_FAILED
	else
	    # Converting to .ppk format with appropriate comment (if any).
	    echo "$passphrase" | $PUTTYGEN "$keyfile" -o "$keyfile.ppk" >/dev/null 2>&1; result=$?
	    if [ $result -ne 0 ]; then log $MSG_SMBM_CNV_FAILED
	    else
		if [ ! -z "$comment" ]; then
		    echo "$passphrase" | $PUTTYGEN -C "$comment" "$keyfile.ppk" >/dev/null 2>&1; result=$?
		fi
		if [ $result -ne 0 ]; then log "$MSG_SMBM_COMMENT_FAILED" "$keyfile.ppk" >/dev/null 2>&1; fi
	    fi
	    # Extracting public key using OpenSSH format.
	    echo "$passphrase" | $PUTTYGEN -L "$keyfile.ppk" >"$keyfile.pub" 2>/dev/null; result=$?
	    if [ $result -ne 0 ]; then log "$MSG_SMBM_CNV_FAILED" "$keyfile.pub"; fi
	    # Adjusting filesystem permissions.
	    chmod 600 "$keyfile" "$keyfile.ppk" >/dev/null 2>&1
	    chmod 644 "$keyfile.pub" >/dev/null 2>&1
	    log "$MSG_SMBM_KEYGEN_SUCCESS $keyfile ($passphrase)"
	fi

    # Fallback to ssh-keygen if necessary.
    else
	if [ -x "$SSHKEYGEN" ]; then
	    # Generating a DSA keypair with AES-128-CBC cipher.
	    rm "$keyfile" "$keyfile.pub" >/dev/null 2>&1
	    $SSHKEYGEN -q -t dsa -N "$passphrase" -C "$comment" -f "$keyfile"; result=$?
	    if [ $result -ne 0 ]; then log "$MSG_SMBM_KEYGEN_FAILED"
	    else
		# Adjusting filesystem permissions (not really necessary).
		chmod 600 "$keyfile" >/dev/null 2>&1
		chmod 644 "$keyfile.pub" >/dev/null 2>&1
		log "$MSG_SMBM_KEYGEN_FALLBACK $keyfile ($passphrase)"
	    fi
	# Fatal if no fallback.
	else
	    log "$MSG_SMBM_KEYGEN_NOKEYGEN"
	fi
    fi
    return
}

# Helper function: logs with timestamp into a logfile and native to STD_ERR.
function log {

    local DATETIME="+%d/%m/%Y:%H:%M:%S %Z"
    # Empty lines are ignored.
    if [ -n "$*" ]; then
	# Into logfile and STD_ERR if exists.
	if [ -f "$SMBMLOGFILE" -a -w "$SMBMLOGFILE" ]; then
    	    echo -e "[$($DATE "$DATETIME")] $($WHOAMI) ${FUNCNAME[1]} $*" >>$SMBMLOGFILE
    	    echo -e "$*" >&2
	# Lack of logfile uses STD_ERR only.
	else
    	    echo -e "[$($DATE "$DATETIME")] $($WHOAMI) ${FUNCNAME[1]} $*" >&2
	fi
    fi
    return
}

# Testing function (in development use only) - intentionally left blank.
function testing {

    return
}

#
# Main program - handling options
#

option=`echo $1 | $SED -n "s/^-\+//p"`; if [ ! -z "$option" ]; then shift; fi
case "$option" in
    "ag"|"addgroup")
        smbm_addgroup "$@"; exit $?
    ;;
    "am"|"aug"|"addmember"|"addusertogroup")
        smbm_addmember "$@"; exit $?
    ;;
    "au"|"adduser")
        smbm_adduser "$@"; exit $?
    ;;
    "configure"|"install")
        smbm_configure "$@"; exit $?
    ;;
    "dg"|"delgroup")
        smbm_delgroup "$@"; exit $?
    ;;
    "dm"|"dug"|"delmember"|"deluserfromgroup")
        smbm_delmember "$@"; exit $?
    ;;
    "getent"|"getentity")
	getentity "$@"; exit $?
    ;;
    "getnextrid")
	getnextrid "$@"; exit $?
    ;;
    "keygen")
	keygen "$@"; exit $?
    ;;
    "lg"|"listgroup")
        smbm_listgroup "$@"; exit $?
    ;;
    "lm"|"lgm"|"listmembers")
        smbm_listmembers "$@"; exit $?
    ;;
    "lms"|"lum"|"lug"|"listmemberships")
        smbm_listmemberships "$@"; exit $?
    ;;
    "lu"|"listuser")
        smbm_listuser "$@"; exit $?
    ;;
    "mg"|"modgroup")
        smbm_modgroup "$@"; exit $?
    ;;
    "mu"|"moduser")
        smbm_moduser "$@"; exit $?
    ;;
    "testing")
	testing "$@"; exit $?
    ;;
    *)
        echo -e "$USAGE"; exit $EXIT_ERR
    ;;
esac

